﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;

namespace TGC.RadioButtonUtil
{

    /// <summary>
    /// Represents a group of related radio buttons.
    /// </summary>
    public class RadioButtonGroup
    {
        private EventHandler checkedChangedHandler = null;
        private EventHandler clickEventHandler = null;  // to handle buttons with AutoCheck == false...
        private EventHandler tabOrderChangedHandler = null;

        private SortedList<int, RadioButton> buttons = new SortedList<int, RadioButton>();
        private RadioButton current = null;

        private bool criticalOpInProgress = false;


        public RadioButtonGroup()
        {
            checkedChangedHandler = new EventHandler(Group_CheckedChanged);
            clickEventHandler = new EventHandler(Button_Click);
            tabOrderChangedHandler = new EventHandler(Group_TabOrderChanged);
        }

        /// <summary>
        /// Adds a RadioButton to the group and selects the first one in the group. 
        /// NOTE: If two radio buttons with the same tab index are added, 
        /// the tab index of the last one is altered (the first larger index  available is assigned).
        /// </summary>
        /// <remarks>This operation might alter the indices of other radio buttons.</remarks>
        public void Add(RadioButton item)
        {
            item.Checked = false;

            while (buttons.ContainsKey(item.TabIndex))
                item.TabIndex++;

            buttons.Add(item.TabIndex, item);

            item.Click += clickEventHandler;    // to handle buttons with AutoCheck == false...
            item.CheckedChanged += checkedChangedHandler;
            item.TabIndexChanged += tabOrderChangedHandler;

            buttons.Values[0].Checked = true;            
        }

        /// <summary>
        /// Removes a RadioButton from the group.
        /// </summary>
        /// <exception cref="System.ArgumentException">
        /// Thrown when <code>item</code> is not a member of the group.
        /// </exception>
        /// <exception cref="System.ArgumentNullException">
        /// Thrown if the argument is a null reference.
        /// </exception>
        /// <remarks>Removal operation might alter the indices of other radio buttons.</remarks>
        public void Remove(RadioButton item)
        {
            if (item == null)
                throw new ArgumentNullException("item");
            
            int index = buttons.IndexOfValue(item);
            if (index == -1)
                throw new ArgumentException("The radio button group does not contain the specified item.");

            item.Click -= clickEventHandler;
            item.CheckedChanged -= checkedChangedHandler;
            item.TabIndexChanged -= tabOrderChangedHandler;

            buttons.RemoveAt(index);

            // If the removed item is the one currently selected, check the first one in the group (if any).
            if (object.ReferenceEquals(item, current))
                if (buttons.Count > 0)
                    buttons.Values[0].Checked = true;

        }

        /// <summary>
        /// Gets the radio button at the given index.
        /// </summary>
        /// <exception cref="System.IndexOutOfRangeException">
        /// Thrown if the given index is invalid.
        /// </exception>
        public RadioButton this[int index]
        {
            get
            {
                if (index < 0 || index >= buttons.Values.Count)
                    throw new IndexOutOfRangeException();

                return buttons.Values[index];
            }
        }

        public int Count
        {
            get { return buttons.Count; }
        }

        private void Group_CheckedChanged(object sender, EventArgs e)
        {
            RadioButton rBtn = (RadioButton)sender;

            if (rBtn.Checked)
            {
                if (current != null && !object.ReferenceEquals(rBtn, current))
                    current.Checked = false;

                current = rBtn;

                OnStateChanged(RadioBtnGroupEventArgs.CreateFromGroup(this));
            }
            else if (current != null && current.Checked == false)
            {
                // The current selection was unchecked (== nothing is selected)

                current = null;
                OnStateChanged(RadioBtnGroupEventArgs.CreateForNoneSelected());
            }
        }

        private void Group_TabOrderChanged(object sender, EventArgs e)
        {
            OnTabOrderChanged();
        }

        private void Button_Click(object sender, EventArgs e)
        {
            RadioButton rBtn = (RadioButton)sender;
            if (rBtn.AutoCheck == false && rBtn.Checked == false)
                rBtn.Checked = true;
        }


        public RadioButton Current
        {
            get { return current; }
        }

        public int CurrentIndex
        {
            get
            {
                if (current == null)
                    throw new InvalidOperationException(
                        "None of the radio buttons is selected! Always check if the Current property is null."
                        );
                
                return buttons.IndexOfValue(current);
            }
        }

        public RadioButton[] Buttons
        {
            get 
            {
                RadioButton[] result = new RadioButton[buttons.Count];

                // A shallow copy...
                buttons.Values.CopyTo(result, 0);
                
                return result; 
            }
        }

        public void DeselectAll()
        {
            // The method uses the fact that only one of the radio buttons can be selected at any time.
            if (current != null)
            {
                current.Checked = false;

                // current will be set to null in the Group_CheckedChanged event handler.
                // The appropriate OnStateChanged() method is called there, too.

                //OnStateChanged(RadioBtnGroupEventArgs.CreateForNoneSelected());
            }
        }

        /// <summary>
        /// State changed event is rased when the state of the group is changed, i.e. when client code
        /// selects an option or when all choices are cleared.
        /// </summary>
        public event EventHandler<RadioBtnGroupEventArgs> StateChanged;

        private void OnStateChanged(RadioBtnGroupEventArgs e)
        {
            // Check if any crytical section of the code is running, or else strange things
            // are bound to happen...

            if (!criticalOpInProgress)
                if (StateChanged != null)
                    StateChanged(this, e);
        }

        /// <summary>
        /// Raised when the tab order of the radio buttons in this group changes. Call Sort() to reorder the radio buttons.
        /// </summary>
        public event EventHandler TabOrderChanged;

        private void OnTabOrderChanged()
        {
            if (TabOrderChanged != null)
                TabOrderChanged(this, new EventArgs());
        }

        /// <summary>
        /// Sorts the radio buttons in this group.
        /// </summary>
        public void Sort()
        {
            criticalOpInProgress = true;

            RadioButton[] temp = new RadioButton[buttons.Count];

            // A shallow copy...
            buttons.Values.CopyTo(temp, 0);

            // Not using buttons.Clear(), because of the handler methods.
            foreach (RadioButton rBtn in temp)
                this.Remove(rBtn);

            foreach (RadioButton rBtn in temp)
                this.Add(rBtn);

            criticalOpInProgress = false;
        }
    }
}
